/*
  Copyright (c) 2022  Kazuhiro Ito (kzhr@d1.dion.ne.jp)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#ifndef __EBEST_COMMON_H__
#define __EBEST_COMMON_H__

#include <config.h>
#include <stdbool.h>
#include <iconv.h>
#include <ebu/eb.h>
#include <ebu/text.h>
#include <ebu/font.h>
#include <ebu/appendix.h>
#include <ebu/error.h>
#include <ebu/binary.h>
#include <estraier.h>

#define EBEST_VERBOSE      "ebest_verbose"
#define EBEST_DEBUG        "ebest_debug"
#define EBEST_DB_TRUNC     "estdb_trunc"
#define EBEST_DB_PERFNG    "estdb_perfng"
#define EBEST_DB_CHRCAT    "estdb_chrcat"
#define EBEST_DB_SMALL     "estdb_small"
#define EBEST_DB_LARGE     "estdb_large"
#define EBEST_DB_HUGE      "estdb_huge"
#define EBEST_DB_HUGE2     "estdb_huge2"
#define EBEST_DB_HUGE3     "estdb_huge3"
#define EBEST_DB_OPTIMIZE  "estdb_optimize"
#define EBEST_DB_OPT_PURGE "estdb_optimize_purge"
#define EBEST_DB_OPT_DB    "estdb_optimize_db"
#define EBEST_DB_CACHE     "estdb_cache"
#define EBEST_DB_CACHE_DEFAULT "64"
#define EBU_GAIJI_ALTERNATE "ebu_gaiji_alternate"
#define EBU_GAIJI_ALTERNATE_DEFAULT " "
#define EBU_CODING_EPWING "ebu_coding_epwing"
#define EBU_CODING_EBXAC  "ebu_coding_ebxac"
#define EBU_CODING_EBG    "ebu_coding_ebg"
#define EBU_USE_GAIJI_MAP "ebu_use_gaiji_map"
#define EBU_CODING_EPWING_DEFAULT "EUC-JP"
#define EBU_CODING_EBXAC_DEFAULT  "EUC-CN"
#define EBU_CODING_EBG_DEFAULT    "ISO-8859-1"

/* kvs.c */
typedef struct {
  unsigned char* key;
  unsigned char* value;
} KVS_Element;

typedef struct {
  int size;
  int msize;
  KVS_Element *list;
} KVS;

KVS* kvs_create();
void kvs_destroy(KVS *kvs);
unsigned char* kvs_get(KVS *kvs, const unsigned char *key);
int kvs_unset(KVS *kvs, const unsigned char *key);
int kvs_set(KVS *kvs, const unsigned char *key, const unsigned char *value);

/* sbuffer.c */
typedef struct {
  size_t size;
  size_t used;
  unsigned char *data;
} SBuffer;

SBuffer *sbuffer_create (size_t size);
void sbuffer_destroy (SBuffer *buffer);
int sbuffer_expand (SBuffer *buffer, size_t step);
int sbuffer_request (SBuffer *buffer, size_t req_size);
int sbuffer_iconv (SBuffer *data, iconv_t iconvd, char *from, size_t length);
int sbuffer_write_char (SBuffer *buffer, int ch);
int sbuffer_write_bytes (SBuffer *buffer, const char *bytes, size_t length);
int sbuffer_write_utf8 (SBuffer *data, const int ch);
SBuffer *sbuffer_load_file(const char *filename);
SBuffer *sbuffer_snprintf(const char* restrict format, ...);
int sbuffer_write_snprintf
(SBuffer *buffer, const char* restrict format, ...);


/* ebu.c */
typedef struct {
  ESTDB *estdb;
  EB_Book *book;
  EB_Appendix *appendix;
  EB_Hookset hookset;
  const char *coding;
  const char *coding_ebxac;
  const unsigned char *gaiji_default;
} EBU_Params;

int ebu_open (int argc, char** argv, int start, EBU_Params *ebu);
int ebu_make_index (EBU_Params *ebu);
void ebu_close (EBU_Params *ebu);

/* utils.c*/
void DEBUG (const char *message, ...);
void VERBOSE (const char *message, ...);
void estdb_informer(const char *message, void *opaque);
char *snprintf_buffer (const char* restrict format, ...);
char *trim_spaces(char *string, bool trim_lf);
void hexdump (FILE *fp, const char *data, int length);


/* variables.c */
int variables_set
(KVS *variables, const unsigned char *key, const unsigned char *value,
 bool system);
int variale_unset(KVS *variables, const unsigned char *key);
const unsigned char *variables_get
(KVS *variables, const unsigned char *key, const unsigned char *def);
bool variables_get_bool(KVS *variables, const unsigned char *key, bool def);
int variagles_set_from_line(KVS *variables, const char *line);


/* ebu_gaiji.c */
SBuffer *ebu_gaiji_map_load(const char *filename);
KVS *ebu_gaiji_map_parse(SBuffer *indata);
KVS *ebu_load_gaiji_map (EB_Book *book);
KVS *ebu_gaiji_register_cp51932(KVS *kvs);

/* ebestmk.c */
extern KVS *variables;
extern FILE *debug;
extern FILE *verbose;
extern bool signaled;

void set_signal_handler();

#endif
