/*
  Copyright (c) 2022  Kazuhiro Ito (kzhr@d1.dion.ne.jp)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

#include <config.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <iconv.h>
#include "common.h"


SBuffer *sbuffer_create (size_t size) {
  if (size == 0) return NULL;
  SBuffer *result = malloc(sizeof(SBuffer));
  if (result) {
    result->data = malloc(size);
    if (!result->data) {
      free(result);
      return NULL;
    }
  }
  result->size = size;
  result->used = 0;
  return result;
}

void sbuffer_destroy (SBuffer *buffer) {
  free(buffer->data);
  free(buffer);
}

int sbuffer_expand (SBuffer *buffer, size_t step) {
  size_t new_size = buffer->size + (step ? step : (buffer->size >> 1));
  unsigned char *new_data = realloc(buffer->data, new_size);
  if (!new_data) return -1;
  buffer->data = new_data;
  buffer->size = new_size;
  return 0;
}

int sbuffer_request (SBuffer *buffer, size_t req_size) {
  if (req_size && (buffer->size - buffer->used) >= req_size) return 0;
  if (req_size > (buffer->size >> 1)) {
    return sbuffer_expand(buffer, req_size);
  } else {
    return sbuffer_expand(buffer, (buffer->size >> 1));
  }
}

int sbuffer_iconv (SBuffer *data, iconv_t iconvd, char *from, size_t length) {
  if (iconvd == (iconv_t) -1)
    return sbuffer_write_bytes(data, from, length);

  iconv(iconvd, NULL, NULL, NULL, NULL);
  while (1) {
    size_t out_left = data->size - data->used;
    char* obuffer = data->data + data->used;
    if (iconv(iconvd, &from, &length, &obuffer, &out_left) != (size_t) -1) {
      data->used = data->size - out_left;
      break;
    } else if (errno == EILSEQ && length > 0) {
      fputs("sbuffer_iconv(): EILSEQ\n", stderr);
      hexdump(stderr, from, (length > 16) ? 16 : length);
      return -1;
    } else if (errno != E2BIG) {
      fprintf(stderr, "iconv failed, %d\n", errno);
      hexdump(stderr, from, (length > 16) ? 16 : length);
      return -1;
    }
    data->used = data->size - out_left;
    if (sbuffer_request(data, data->size >> 1)) {
      fprintf(stderr, "Memory exhausted\n");
      return -1;
    }
  }
  return 0;
}

int sbuffer_write_char (SBuffer *buffer, int ch) {
  int result = sbuffer_request(buffer, 1);

  if (result) return result;
  buffer->data[buffer->used] = (unsigned char) ch;
  buffer->used++;

  return 0;
}

int sbuffer_write_bytes (SBuffer *buffer, const char *bytes, size_t length) {
  int result = sbuffer_request(buffer, length);

  if (result) return result;
  memcpy (buffer->data + buffer->used, bytes, length);
  buffer->used += length;

  return 0;
}

int sbuffer_write_utf8 (SBuffer *data, const int ch) {
  unsigned char *pos = data->data + data->used;
  int result;
  
  if (ch < 0x80) {
    if (result = sbuffer_request(data, 1)) return result;
    *pos = (unsigned char) ch;
    data->used++;
  } else if (ch < 0x800) {
    if (result = sbuffer_request(data, 2)) return result;
    *(pos    ) = (unsigned char)(( ch >>  6)         | 0xc0);
    *(pos + 1) = (unsigned char)(( ch        & 0x3f) | 0x80);
    data->used += 2;
  } else if (ch < 0x10000) {
    if (result = sbuffer_request(data, 3)) return result;
    *(pos    ) = (unsigned char)(( ch >> 12)         | 0xe0);
    *(pos + 1) = (unsigned char)(((ch >>  6) & 0x3f) | 0x80);
    *(pos + 2) = (unsigned char)(( ch        & 0x3f) | 0x80);
    data->used += 3;
  } else {
    if (result = sbuffer_request(data, 4)) return result;
    *(pos    ) = (unsigned char)(( ch >> 18)         | 0xf0);
    *(pos + 1) = (unsigned char)(((ch >> 12) & 0x3f) | 0x80);
    *(pos + 2) = (unsigned char)(((ch >>  6) & 0x3f) | 0x80);
    *(pos + 3) = (unsigned char)(( ch        & 0x3f) | 0x80);
    data->used += 4;
  }

  return 0;
}

SBuffer *sbuffer_load_file(const char *filename) {
  FILE *infile = NULL;
  SBuffer *filedata = sbuffer_create (4096);
  size_t read;
  
  if (!filedata) {
    fprintf(stderr, "Memory exhausted.\n");
    goto error;
  }

  infile = fopen(filename, "rb");
  if (!infile) {
    fprintf(stderr, "failed to open file, %s.\n", filename);
    goto error;
  }

  while (1) {
    read = fread(filedata->data + filedata->used,
		 1, filedata->size - filedata->used, infile);
    filedata->used += read;
    if (feof(infile)) {
      break;
    } else if (filedata->used == filedata->size) {
      if (sbuffer_request(filedata, 0)) {
	fprintf(stderr, "Memory exhausted.\n");
	goto error;
      }
    } else {
      fprintf(stderr, "file I/O error.\n");
      goto error;
    }
  }
  fclose(infile);
  
  if (sbuffer_write_char(filedata, 0))
    goto error;
  
  return filedata;
 error:
  if (filedata) sbuffer_destroy(filedata);
  if (infile) fclose(infile);
  return NULL;
}

SBuffer *sbuffer_snprintf(const char* restrict format, ...) {
  va_list ap;
  SBuffer *result = NULL;

  va_start(ap, format);
  int size = vsnprintf(NULL, 0, format, ap);
  va_end(ap);
  if (size < 0) goto snprintf_error;

  result = sbuffer_create((size_t)size + 1);
  if (!result) {
    fprintf(stderr, "Memory exhausted\n");
    goto error;
  }

  va_start(ap, format);
  int size2 = vsnprintf(result->data, size + 1, format, ap);
  va_end(ap);
  if ( size != size2)
    goto snprintf_error;

  return result;

 snprintf_error:
  fprintf(stderr, "vsnprintf() error\n");
 error:
  if (result) sbuffer_destroy(result);
  return NULL;
}

int sbuffer_write_snprintf
(SBuffer *buffer, const char* restrict format, ...) {
  va_list ap;

  va_start(ap, format);
  int size = vsnprintf(NULL, 0, format, ap);
  va_end(ap);
  if (size < 0) goto snprintf_error;

  if (sbuffer_request(buffer, size + 1)) {
    fprintf(stderr, "Memory exhausted\n");
    goto error;
  }

  va_start(ap, format);
  int size2 = vsnprintf(buffer->data + buffer->used, size + 1, format, ap);
  va_end(ap);
  if (size != size2) {
    goto snprintf_error;
  }
  buffer->used += size + 1;

  return size;

 snprintf_error:
  fprintf(stderr, "vsnprintf() error\n");
 error:
  return -1;
}
